module Typing (
    mgt, -- most general type with respect to a given environment
    infer
  ) where
import CoreFP
import Unification

-- first environment for constants, second one for variables
type IP = (Env, Env, Exp, Type)

maxIdx :: Type -> Int
maxIdx (TVar i)      =  i
maxIdx (TFun t1 t2)  =  max (maxIdx t1) (maxIdx t2)
maxIdx (TCon _ ts)   =  maximum (0 : map maxIdx ts)

incBy :: Int -> Type -> Type
incBy i (TVar j)      =  TVar (i + j)
incBy i (TFun t1 t2)  =  TFun (incBy i t1) (incBy i t2)
incBy i (TCon g ts)   =  TCon g (map (incBy i) ts)

typeOf :: String -> Env -> Type
typeOf id env =
  case lookup id env of
    Just t  -> t
    Nothing -> error ("no type for '" ++ id ++ "'")

toUp :: IP -> UP
toUp ip@(_, _, _, t) = step (maxIdx t + 1) [] [ip]
  where
    step i up []                  =  up
    step i up ((cs, xs, e, t):ip) =
      case e of
        Var x        -> step i ((typeOf x xs, t):up) ip
        Con c        -> let t' = typeOf c cs in
          step (i + maxIdx t' + 1) ((incBy i t', t):up) ip
        App e1 e2    -> step (i+1) up
          ((cs, xs, e1, TVar i @-> t):(cs, xs, e2, TVar i):ip)
        Abs x e      -> step (i+2) ((t, TVar i @-> TVar (i+1)):up)
          ((cs, (x, TVar i):xs, e, TVar (i+1)):ip)
        Let x e1 e2  -> step (i+1) up
          ((cs, xs, e1, TVar i):(cs, (x, TVar i):xs, e2, t):ip)
        Ite e1 e2 e3 -> step i up
          ((cs, xs, e1, tbool):(cs, xs, e2, t):(cs, xs, e3, t):ip)

mgt :: Env -> Env -> Exp -> Type -> Type
mgt cs xs e t = t `tsub` mgu
  where mgu = unify (toUp (cs, xs, e, t))

infer :: String -> Type
infer s = mgt primitives [] e (TVar (-1))
  where e = fromString s
